import codecs


def lookup(coding_name: str) -> codecs.CodecInfo | None:
    """Function registered with the codecs module to find the bleach codec."""

    if coding_name != "bleach":
        return None

    return codecs.CodecInfo(
        encode=encode_bleach,  # type: ignore
        decode=decode_bleach,  # type: ignore
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


def encode_bleach(input_chars: str, errors: str = "strict") -> tuple[bytes, int]:
    """Bleach a file, replacing everything with spaces and tabs.

    Converts the file into a set of UTF-8 encoded bytes, then generates a
    string representation of the bytes using spaces for 1s and tabs for zeros.
    This encoding also writes a newline every 40 characters, to avoid
    extremely long line lengths. The newlines could be removed without affecting
    the encoded document.
    """

    bleached_lines = []
    byte_encoding = input_chars.encode("UTF-8")
    as_bits = "".join([f"{x:08b}" for x in byte_encoding])
    bleached = as_bits.replace("1", " ").replace("0", "\t")
    for i in range(0, len(bleached), 40):
        bleached_lines.append(bleached[i : i + 40])

    # Terminate with blank line.
    bleached_lines.append("")

    # We're only using tabs and spaces, we know ascii is sufficient.
    return "\n".join(bleached_lines).encode("ascii"), len(input_chars)


def decode_bleach(input_bytes: bytes, errors: str = "strict") -> tuple[str, int]:
    """Unbleach a file by decoding tabs and spaces into bytes.

    Replaces all spaces with 1s and tabs with 0s, and interprets each chunk of
    8 characters as a byte. The resulting document is a series of UTF-8
    encoded bytes representing the original document. Newlines are ignored.
    """
    input_lines = bytearray(input_bytes).decode("ascii").split("\n")
    output_bytes = bytearray()

    for line in input_lines:
        if line.startswith("#"):
            continue

        binary_str = line.replace(" ", "1").replace("\t", "0")
        for i in range(0, len(binary_str), 8):
            b = int(binary_str[i : i + 8], 2)
            output_bytes.append(b)

    return output_bytes.decode("UTF-8"), len(input_bytes)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def encode(self, input: str, final: bool = False) -> bytes:
        return encode_bleach(input, self.errors)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):
    def __init__(self, errors: str = "strict") -> None:
        super().__init__(errors)
        self.buffer = bytearray()
        self.i = 0

    def decode(self, input: bytes, final: bool = False) -> str:  # type: ignore
        if not final:
            self.buffer.extend(input)
            return ""

        result = bytes(self.buffer)
        # Needs to make sure this returns "" if called again with final=True.
        # If it returns output, this will get called again and again.
        self.buffer.clear()
        if result:
            # The interpreter chomps the first line from the decoded text, so
            # we add a comment here.
            return "# coding=utf-8\n" + decode_bleach(result)[0]

        return ""


class StreamWriter(codecs.StreamWriter):
    encode = encode_bleach  # type: ignore


class StreamReader(codecs.StreamReader):
    decode = decode_bleach  # type: ignore


# Register codec when this file is imported.
codecs.register(lookup)



import pathlib
import runpy
import site
import typing

import click

import acme.bleach.codec  # noqa: F401


PTH_CONTENTS = """\
# Provides the bleach codec to scripts
import acme.bleach.codec
"""


@click.group
def main() -> None:
    """Remove unsightly visible characters from your source code.

    Replaces the unsightly visible characters in Python source code with strings
    of spaces and tabs.
    """
    pass


@main.command()
@click.argument("infile", type=click.File("r"))
@click.option(
    "-o", "--outfile", type=click.File("wb"), help="Write to this file.", default="-"
)
def bleach(infile: typing.TextIO, outfile: typing.BinaryIO) -> None:
    """Bleach a file by replacing unsightly visible characters.

    Writes to standard output by default.

    INFILE: Input file to bleach.
    """
    bleached = infile.read().encode("bleach")

    outfile.write(b"# coding=bleach\n")
    outfile.write(bleached)


@main.command()
@click.argument("infile", type=click.File("rb"))
@click.option(
    "-o", "--outfile", type=click.File("w"), help="Write to this file.", default="-"
)
def unbleach(infile: typing.TextIO, outfile: typing.BinaryIO) -> None:
    """Unbleach a bleached file, restoring those visible characters.

    Writes to standard output by default.

    INFILE: Input file to unbleach.
    """
    unbleached = infile.read().decode("bleach")  # type: ignore
    outfile.write(unbleached)


@main.command()
@click.option("-m", "--module", type=str, help="Run library module as script")
@click.argument(
    "filename", type=click.Path(exists=True, dir_okay=False), required=False
)
def run(filename: str, module: str | None) -> None:
    """Run a bleached (or unbleached) script.

    Wrapper around "python script.py" or "python -mscript" that understands
    "# coding=bleach" comments. If the "install" command has been run, this
    wrapper is not necessary.

    FILENAME: Script to run.
    """

    if module:
        runpy.run_module(module)
    else:
        runpy.run_path(filename)


def _try_write_pth_to(directory: str) -> None:
    """Try and write pth file, propagating IOErrors."""
    path = pathlib.Path(directory, "bleach.pth")
    with path.open("w") as pth_out:
        pth_out.write(PTH_CONTENTS)
    print(f"Wrote {path}.")


@main.command()
@click.option("-u/-s", "--user/--system")
def install(user: bool | None) -> None:
    """Install the bleach codec to be used natively.

    This command will install a bleach.pth file into your Python installation's
    site-packages folder, enabling you to run bleached files with
    "# coding=bleach" without having to use "bleach run".
    """
    # Can't write to the user site if it doesn't exit.
    if user and not site.ENABLE_USER_SITE:
        raise click.ClickException(
            "Option --user was used to force use of user-site packages, "
            "but the user site is disabled."
        )

    # Prefer the user site over the system-wide site if not explicitly set.
    if user or (user is None and site.ENABLE_USER_SITE):
        try:
            _try_write_pth_to(site.getusersitepackages())
        except OSError as e:
            raise click.ClickException(
                f"Could not write to user site-packages. {e}"
            ) from e

    # Write to any system package if possible.
    else:
        for site_package in site.getsitepackages():
            try:
                _try_write_pth_to(site_package)
                return
            except OSError:
                pass

        raise click.ClickException(
            "Insufficient permissions to write to any of " f"{site.getsitepackages()}."
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['acme', 'acme.bleach']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1,<9.0']

entry_points = \
{'console_scripts': ['bleach = acme.bleach.app:main']}

setup_kwargs = {
    'name': 'acme-bleach',
    'version': '0.2.3',
    'description': 'Remove unsightly non-whitespace characters from your code.',
    'long_description': '# Bleach\n\n[![Read the documentation at https://acme-bleach.readthedocs.io/](https://img.shields.io/readthedocs/acme-bleach/latest.svg?label=Read%20the%20Docs)][read the docs]\n\n[read the docs]: https://acme-bleach.readthedocs.io/\n[tests]: https://github.com/stefansm/acme-bleach/actions?workflow=Tests\n\nRemove unsightly visible characters from your code. Inspired by the [`Acme::Bleach` Perl module][acme::bleach] by Damian\nConway.\n\n[acme::bleach]: https://metacpan.org/pod/Acme::Bleach\n\n## Installation\n\nYou can install _Bleach_ via [pip] from [PyPI]:\n\n```console\n$ pip install acme-bleach\n```\n\n## Usage\n\nLet\'s start with the following Python script, `hello.py`:\n\n```python\nprint("Hello, world!")\n```\n\n### Bleaching\n\nTo bleach it, run the following command:\n\n```console\n$ bleach bleach hello.py -o hello-bleached.py\n```\n\nThe result is a bleached version of the original script. The contents of\n`hello-bleached.py` are (representing tabs with `»»»»` and spaces with `·`):\n\n```\n# coding=bleach\n»»»»···»»»»»»»»»»»»»»»»»···»»»»»·»»»»»»»··»»·»»»»»»»·»»»··»»···»»»»»···»·»»»»»»»\n»»»»»»»»·»»»·»»»»»»»»»»»»»»»»»»»·»»»»»»»»»»»·»»»»»»»·»»»»»»»·»»»»»»»»»»»»»»»··»»»»»»·»»»·»»»··»»··»»»»»»\n»»»»··»»··»»»»»»»»»»··»»····»»»»»»»»·»»»··»»»»»»»»»»»»»»·»»»»»»»»»»»»»»»»»»»»»»»···»···\n»»»»··»»····»»»»···»»»»»·»»»»»»»··»»··»»»»»»»»»»··»»»»»»·»»»»»»»»»»»»»»»·»»»»»»»»»»»»»»»·\n»»»»»»»»·»»»»»»»»»»»·»»»»»»»»»»»·»»»·»»»»»»»·»»»»»»»»»»»»»»»·»»»·»»»\n```\n\n### Running\n\nYou can run the bleached file using `bleach run`:\n\n```console\n$ bleach run hello-bleached.py\nHello, world!\n```\n\n### Unbleaching\n\nYou can unbleach the file using `bleach unbleach`:\n\n```console\n$ bleach unbleach hello-bleached.py -o hello-unbleached.py\n$ cat hello-unbleached.py\nprint("Hello, world!")\n```\n\n### Installing the bleach codec\n\nRunning scripts with `bleach run` is far too much effort. If you want to run bleached scripts directly, you can install\nthe `bleach` codec:\n\n```console\n$ bleach install\nWrote /home/stefans/bleach/.venv/lib/python3.10/site-packages/bleach.pth\n$ python hello-bleached.py\nHello, world!\n```\n\nInstalling the codec will attempt to write a `.pth` file to the `site-packages` directory of the current Python\ninterpreter. Note the `import` in the `.pth` file will be evaluated every time the current Python runs, so there is a\ntiny amount of overhead.\n\n## License\n\nDistributed under the terms of the [MIT license][license],\n_Bleach_ is free and open source software.\n\n## Credits\n\nThis project was generated from a variant of [@cjolowicz]\'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[pypi]: https://pypi.org/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[pip]: https://pip.pypa.io/\n\n<!-- github-only -->\n\n[license]: https://github.com/stefansm/acme-bleach/blob/main/LICENSE\n[command-line reference]: https://acme-bleach.readthedocs.io/en/latest/usage.html\n',
    'author': 'Stefans Mezulis',
    'author_email': 'stefans.mezulis@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/stefansm/acme-bleach',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
